! $Id: PolyScalarSharedPointerExample.f90 2220 2016-06-23 00:17:02Z ian $
! Copyright 2016 M.E.G.M.S.  See LICENCE.txt for licence (Apache 2.0).
! See http://www.megms.com.au/shared-pointers.htm for more information.

! Example and simple tests of the code in PolyScalarSharedPointer.i90.

! A module that defines the type that the shared pointer will 
! reference.  In this example, the dynamic type of the referenced 
! object will be an extension of the object type that the 
! shared pointer references.

MODULE MyTypes
  IMPLICIT NONE
  
  ! The parent type.  This will be used as the object_type
  TYPE, ABSTRACT, PUBLIC :: MyParent
  CONTAINS
    PROCEDURE(parent_Binding), DEFERRED :: Binding
  END TYPE MyParent
  
  ABSTRACT INTERFACE
    SUBROUTINE parent_Binding(arg)
      IMPORT :: MyParent
      CLASS(MyParent), INTENT(IN) :: arg
    END SUBROUTINE parent_Binding
  END INTERFACE
  
  ! An extension of MyParent, that will be used as the dynamic 
  ! type of the object that the shared pointer references.
  TYPE, EXTENDS(MyParent), PUBLIC :: MyExtension
    CHARACTER(:), ALLOCATABLE :: comp
  CONTAINS
    PROCEDURE :: Binding => extension_Binding
    FINAL :: extension_Final
  END TYPE MyExtension
CONTAINS
  SUBROUTINE extension_Binding(arg)
    CLASS(MyExtension), INTENT(IN) :: arg
    PRINT "(2X,'Binding for ',A)", arg%comp
  END SUBROUTINE extension_Binding
  
  IMPURE ELEMENTAL SUBROUTINE extension_Final(arg)
    TYPE(MyExtension), INTENT(INOUT) :: arg
    PRINT "(2X,'The object for ',A,' is going away.')", arg%comp
  END SUBROUTINE extension_Final
END MODULE MyTypes


! A module that incorporates the shared pointer template code, 
! to define a type SharedPointer and some associated generic 
! procedures for working with such pointers.

MODULE MySharedPointer
  USE MyTypes, object_type => MyParent
  IMPLICIT NONE
  PRIVATE
  
  ! The included source does not make anything explicitly 
  ! PUBLIC, hence the following.
  PUBLIC :: SharedPointer
  PUBLIC :: Acquire
  PUBLIC :: Get
  PUBLIC :: IsAssociated
  
  INCLUDE 'PolyScalarSharedPointer.i90'
END MODULE MySharedPointer


! Example/test driver.

PROGRAM PolyScalarExample
  USE MyTypes
  USE MySharedPointer
  IMPLICIT NONE
  
  ! Procedures that demonstrate/test use of the shared pointer.
  CALL test01
  CALL test02
  CALL test03
  CALL test04
  CALL test05
  
  ! These procedures show how you can break the shared pointer.
!  CALL evil01
!  CALL evil02
CONTAINS
  SUBROUTINE test01
    TYPE(SharedPointer) :: a, b
    
    PRINT "('test01')"
    
    ! Before definition, a shared pointer references no object.
    PRINT "(2X,'Does a reference an object: ',L1)", IsAssociated(a)
    
    BLOCK
      CLASS(MyParent), POINTER :: object
      ! Set up the initial object.
      PRINT "(2X,'Creating Red')"
      ALLOCATE(object, SOURCE=MyExtension('Red'))
      ! Hand control of the lifetime of that object to a shared pointer.
      CALL Acquire(a, object)
      
      ! After calling Acquire, the lifetime of the target of the 
      ! Fortran pointer is managed by the shared pointer.  Consistent 
      ! with this, the association status of the Fortran pointer is 
      ! set to dissociated.
      PRINT "(2X,'The association status of object is ',L1)",  &
          ASSOCIATED(object)
      PRINT "(2X,'Does a reference an object: ',L1)",  &
          IsAssociated(a)
    END BLOCK
    
    
    ! The Item component can be used to work with the referenced 
    ! object.  This should print `Red`.
    CALL a%Item%Binding()
    
    ! Assignment can be used to create additional references to 
    ! the same object.
    b = a
    CALL b%Item%Binding()
    
    ! A shared pointer can be reset by assigning an empty pointer 
    ! to it.
    a = SharedPointer()
    
    ! The object that was referenced by `a` persists, because 
    ! `b` also references it.
    !
    ! The `b` shared pointer still references the object, but when 
    ! `b` goes out of scope at the end of the procedure the object 
    ! will be deallocated.
    PRINT "(2X,'end test01')"
  END SUBROUTINE test01
  
  SUBROUTINE test02
    TYPE(SharedPointer) :: a, b
    CLASS(MyParent), POINTER :: p
    
    PRINT "('test02')"
    
    BLOCK
      CLASS(MyParent), POINTER :: object
      ! Set up the initial object.
      PRINT "(2X,'Creating Blue')"
      ALLOCATE(object, SOURCE=MyExtension('Blue'))
      ! Hand control of the lifetime of that object to a shared pointer.  
      ! This time use the overloaded structure constructor.
      a = SharedPointer(object)
    END BLOCK
    
    p => Get(a)
    CALL p%Binding()
    
    BLOCK
      CLASS(MyParent), POINTER :: object
      ! Set up the initial object.
      PRINT "(2X,'Creating Green')"
      ALLOCATE(object, SOURCE=MyExtension('Green'))
      ! Hand control of the lifetime of that object to a shared pointer.
      b = SharedPointer(object)
    END BLOCK
    
    ! Green will not be referenced by anything after this assignment, 
    ! so it will be destroyed.  a will reference Red.
    a = b
    
    p => Get(a)
    CALL p%Binding()
    
    PRINT "(2X,'end test02')"
  END SUBROUTINE test02
  
  ! Like test02, but with shared pointer inside other type.
  SUBROUTINE test03
    TYPE :: t
      TYPE(SharedPointer) :: sp_comp
    END TYPE t
    TYPE(t) :: x, y
    CLASS(MyParent), POINTER :: p
    
    PRINT "('test03')"
    
    ! Point x%sp_comp at Blue.
    x%sp_comp = make_ptr('Blue')
    
    p => Get(x%sp_comp)
    CALL p%Binding()
    
    ! Point y%sp_comp at Green.
    BLOCK
      CLASS(MyParent), POINTER :: object
      PRINT "(2X,'Creating Green')"
      ALLOCATE(object, SOURCE=MyExtension('Green'))
      y%sp_comp = SharedPointer(object)
    END BLOCK
    
    x = y
    
    p => Get(x%sp_comp)
    CALL p%Binding()
    PRINT "(2X,'end test03')"
  END SUBROUTINE test03
  
  SUBROUTINE test04
    TYPE(SharedPointer) :: a
    
    PRINT "('test04')"
    
    a = make_ptr('Blue')
    
    PRINT "(2X,'Before worker')"
    CALL test04_worker(a)
    PRINT "(2X,'After worker')"
    PRINT "(2X,'end test04')"
  END SUBROUTINE test04
  
  SUBROUTINE test04_worker(x)
    TYPE(SharedPointer), VALUE :: x
    
    ! arg is not finalized.
  END SUBROUTINE test04_worker
  
  SUBROUTINE test05
    TYPE(SharedPointer) :: a
    TYPE(SharedPointer), ALLOCATABLE :: b
    
    PRINT "('test05')"
    
    a = make_ptr('Blue')
    
    ! This is intrinsic assignment.
    b = a
    
    a = SharedPointer()
    
    DEALLOCATE(b)
    
    PRINT "(2X,'end test05')"
  END SUBROUTINE test05
  
  SUBROUTINE evil01
    TYPE(SharedPointer) :: a
    
    PRINT "('evil01')"
    
    a = make_ptr('Blue')
    
    PRINT "(2X,'Before worker')"
    CALL evil01_worker(a)
    PRINT "(2X,'After worker')"
    PRINT "(2X,'end evil01')"
  END SUBROUTINE evil01
  
  SUBROUTINE evil01_worker(x)
    TYPE(SharedPointer), VALUE :: x
    
    ! Vary x by pointing it at another object.
    x = make_ptr('Green')
    
    ! x not finalized here - memory leak.
  END SUBROUTINE evil01_worker
  
  SUBROUTINE evil02
    TYPE(SharedPointer) :: a
    TYPE(SharedPointer), ALLOCATABLE :: b
    
    PRINT "('evil02')"
    
    a = make_ptr('Blue')
    
    ! Does not update reference count.
    ALLOCATE(b, SOURCE=a)
    
    ! Reference count will be zero - object and control block deleted.
    a = SharedPointer()
    
    ! Finalization of b will attempt to reference a control block
    ! that has already gone.
    PRINT "(2X,'end evil02')"
  END SUBROUTINE evil02
  
  FUNCTION make_ptr(colour) RESULT(ptr)
    CHARACTER(*), INTENT(IN) :: colour
    TYPE(SharedPointer) :: ptr
    CLASS(MyParent), POINTER :: object
    
    PRINT "(2X,'Creating ',A)", colour
    ALLOCATE(object, SOURCE=MyExtension(colour))
    ptr = SharedPointer(object)
  END FUNCTION make_ptr
END PROGRAM PolyScalarExample

